"""Functions for comparing output across models.

- thread_multigen(), query_multigen() and multigen(): Compare outputs across models.
- cycle_gen_print(): For a list of models, sequentially grow a Thread with model responses to given IN messages.
"""

from typing import Any, Optional, Union, Callable

import json, csv, re
from io import StringIO

import logging
logger = logging.getLogger(__name__)


from .gen import (
    GenConf,
    GenOut
)

from .thread import (
    Thread,
    MsgKind
)

from .context import (
    Trim,
    Context
)

from .model import (
    Model
)

from .json_schema import JSchemaConf

from .models import Models



def _default_gencall_text(model: Model,
                          thread: Thread,
                          genconf: Optional[GenConf] = None) -> GenOut:
    out = model.gen(thread, genconf)
    return out



def make_dataclass_gencall(cls: Any, # dataclass definition
                           schemaconf: Optional[JSchemaConf] = None) -> Callable:

    def dataclass_gencall(model: Model,
                          thread: Thread,
                          genconf: Optional[GenConf] = None) -> GenOut:    
        out = model.gen_dataclass(cls, 
                                  thread,
                                  genconf,
                                  schemaconf)
        return out

    return dataclass_gencall


def make_pydantic_gencall(cls: Any, # Pydantic BaseModel class
                          schemaconf: Optional[JSchemaConf] = None) -> Callable:

    def pydantic_gencall(model: Model,
                         thread: Thread,
                         genconf: Optional[GenConf] = None) -> GenOut:    
        out = model.gen_pydantic(cls, 
                                 thread,
                                 genconf,
                                 schemaconf)
        return out

    return pydantic_gencall



def make_extract_gencall(target: Any,
                         schemaconf: Optional[JSchemaConf] = None) -> Callable:

    def extract_gencall(model: Model,
                        thread: Thread,
                        genconf: Optional[GenConf] = None) -> GenOut:    
        out = model.gen_extract(target, 
                                thread,
                                genconf,
                                schemaconf)
        return out

    return extract_gencall







def multigen(threads: list[Thread],
             *,
             models: Optional[list[Model]] = None, # existing models
                
             model_names: Optional[list[str]] = None,
             model_names_del_after: bool = True,
    
             gencall: Optional[Callable] = None,
             genconf: Optional[GenConf] = None
             ) -> list[list[GenOut]]:
    """Generate a list of Threads in multiple models, returning the GenOut for each [thread,model] combination.
    
    Actual generation for each model is implemented by the gencall arg Callable with this signature:
        def gencall(model: Model,
                    thread: Thread,
                    genconf: GenConf) -> GenOut
    
    Args:
        threads: List of threads to input into each model.
        models: A list of initialized models. Defaults to None.
        model_names: --Or-- A list of Models names. Defaults to None.
        model_names_del_after: Delete model_names models after using them: important or an out-of-memory error will eventually happen. Defaults to True.
        gencall: Callable function that does the actual generation. Defaults to None, which will use a text generation default function.
        genconf: Model generation configuration to use in models. Defaults to None, meaning default values.

    Raises:
        ValueError: Only one of models or model_names can be given.

    Returns:
        A list of lists in the format [thread,model] of shape (len(threads), len(models)). For example: out[0] holds threads[0] results on all models, out[1]: threads[1] on all models, ...
    """

    if not ((models is None) ^ ((model_names is None))):
        raise ValueError("Only one of models or model_names can be given")

    if gencall is None:
        gencall = _default_gencall_text
    
    mod_count = len(models) if models is not None else len(model_names) # type: ignore[arg-type]

    all_out = []
    
    for i in range(mod_count):
        if models is not None:
            model = models[i]
            logger.debug(f"Model: {model.desc}")
        else:
            name = model_names[i] # type: ignore[index]
            model = Models.create(name)
            logger.info(f"Model: {name} -> {model.desc}")

        mod_out = []
        for th in threads:
            out = gencall(model, th, genconf)
    
            mod_out.append(out)

        all_out.append(mod_out)

        if model_names_del_after and models is None:
            del model
        
    # all_out is currently shaped (M,T) -> transpose to (T,M), so that each row contains thread t for all models
    tout = []
    for t in range(len(threads)):
        tmout = [] # thread t for all models
        for m in range(mod_count):
            tmout.append(all_out[m][t])

        tout.append(tmout)
        
    return tout




def nice_print(type: str,
               val: Any,
               json_kwargs: dict
               ):
    
    if type == "dic":
        text = json.dumps(val, **json_kwargs)
        
    elif type == "value":
        indent_text = " " * 4
        text = re.sub(r", ([a-z0-9_]+)=", 
                      ",\n" + indent_text + "\\1=", 
                      repr(val),
                      flags=re.IGNORECASE)
        
    else:
        text = str(val)

    return text



def format_text(f: StringIO,
                table: list[list[GenOut]], # [ins, model_outs_for_in]
               
                title_list: list[str],
                model_names: list[str],
                   
                out_keys: list[str] = ["text","dic","value"],
        
                json_kwargs: dict = {"indent": 2,
                                     "sort_keys": False,
                                     "ensure_ascii": False
                                     }):

    def lprint(outs_for_in, model_names):
        
        for index,out in enumerate(outs_for_in): # foreach model out
            print("=" * 20, model_names[index], "->", out.res.name, file=f)

            wrote_anything = False
            first = True
            
            out_dict = out.as_dict()
            for k in out_keys:
                
                if k in out_dict and out_dict[k] is not None:
                    
                    if not first and wrote_anything:
                        print("-" * 20, file=f)
                    else:
                        first = False

                    text = nice_print(k, out_dict[k], json_kwargs)
                    print(text, file=f)

                    if out_dict[k]:
                        wrote_anything = True
                        
            if not wrote_anything:        
                text = nice_print("text", out.text, json_kwargs)
                print(text, file=f)
                    
            
    for il in range(len(title_list)): # for each in
        print("/" * 60, "\n", 
              title_list[il], "\n",
              "/" * 60,
              sep="",
              file=f
              )
        lprint(table[il], model_names)
        print(file=f)




def format_csv(f: StringIO,
               table: list[list[GenOut]], # [ins, model_outs_for_in]
               
               title_list: list[str],
               model_names: list[str],
                   
               out_keys: list[str] = ["text","dic","value"],
        
               json_kwargs: dict = {"indent": 2,
                                    "sort_keys": False,
                                    "ensure_ascii": False
                                    }):
    fieldnames = ["Threads"] + model_names
    
    writer = csv.writer(f, 
                        dialect=csv.excel, 
                        )
    
    writer.writerow(fieldnames)
            
    for il in range(len(title_list)): # for each in

        row = [title_list[il]]

        for index,out in enumerate(table[il]): # foreach model out

            cell = ""
            out_dict = out.as_dict()
            first = True
            for k in out_keys:
                
                if k in out_dict and out_dict[k] is not None:
                    
                    if not first and cell:
                        cell += "\n" + "-" * 20 + "\n"
                    else:
                        first = False

                    text = nice_print(k, out_dict[k], json_kwargs)
                    cell += text
                        
            if not cell:
                cell = nice_print("text", out.text, json_kwargs)

            row.append(cell)
            
        writer.writerow(row)






def thread_multigen(threads: list[Thread],
                    model_names: list[str],
                   
                    text: Union[str,list[str],None] = None,
                    csv: Union[str,list[str],None] = None,
                   
                    gencall: Optional[Callable] = None,                   
                    genconf: Optional[GenConf] = None,
    
                    out_keys: list[str] = ["text","dic", "value"],
                   
                    thread_titles: Optional[list[str]] = None                   
                    ) -> list[list[GenOut]]:
    """Generate a single thread on a list of models, returning/saving results in text/CSV.

    Actual generation for each model is implemented by an optional Callable with this signature:
        def gencall(model: Model,
                    thread: Thread,
                    genconf: GenConf) -> GenOut

    Args:
        threads: List of threads to input into each model.
        model_names: A list of Models names.
        text: An str list with "print"=print results, path=a path to output a text file with results. Defaults to None.
        csv: An str list with "print"=print CSV results, path=a path to output a CSV file with results. Defaults to None.
        gencall: Callable function that does the actual generation. Defaults to None, which will use a text generation default function.
        genconf: Model generation configuration to use in models. Defaults to None, meaning default values.
        out_keys: A list with GenOut members to output. Defaults to ["text","dic", "value"].
        thread_titles: A human-friendly title for each Thread. Defaults to None.

    Returns:
        A list of lists in the format [thread,model] of shape (len(threads), len(models)). For example: out[0] holds threads[0] results on all models, out[1]: threads[1] on all models, ...
    """

    assert isinstance(model_names, list), "model_names must be a list of strings"
    
    table = multigen(threads,
                     model_names=model_names, 
                     gencall=gencall,
                     genconf=genconf)
    
    # table[threads,models]

    if thread_titles is None:
        thread_titles = [str(th) for th in threads]

    def format(format_fn, cmds):
        if cmds is None or not cmds:
            return

        f = StringIO(newline='')

        format_fn(f,
                  table, 
                  title_list=thread_titles,
                  model_names=model_names,
                  out_keys=out_keys)
        fmtd = f.getvalue()
        
        if not isinstance(cmds, list):
            cmds = [cmds]
        for c in cmds:
            if c == 'print':
                print(fmtd)
            else: # path
                with open(c, "w", encoding="utf-8") as f:
                    f.write(fmtd)
                
    format(format_text, text)
    format(format_csv, csv)
        
    return table







def query_multigen(in_list: list[str],
                   inst_text: str,                                                
                   model_names: list[str],
   
                   text: Union[str,list[str],None] = None, # "print", path
                   csv: Union[str,list[str],None] = None, # "print", path
                
                   gencall: Optional[Callable] = None,                   
                   genconf: Optional[GenConf] = None,
    
                   out_keys: list[str] = ["text","dic", "value"],
                   in_titles: Optional[list[str]] = None
                   ) -> list[list[GenOut]]:
    """Generate an INST+IN thread on a list of models, returning/saving results in text/CSV.

    Actual generation for each model is implemented by an optional Callable with this signature:
        def gencall(model: Model,
                    thread: Thread,
                    genconf: GenConf) -> GenOut

    Args:
        in_list: List of IN messages to initialize Threads.
        inst_text: The common INST to use in all models.
        model_names: A list of Models names.
        text: An str list with "print"=print results, path=a path to output a text file with results. Defaults to None.
        csv: An str list with "print"=print CSV results, path=a path to output a CSV file with results. Defaults to None.
        gencall: Callable function that does the actual generation. Defaults to None, which will use a text generation default function.
        genconf: Model generation configuration to use in models. Defaults to None, meaning default values.
        out_keys: A list with GenOut members to output. Defaults to ["text","dic", "value"].
        in_titles: A human-friendly title for each Thread. Defaults to None.

    Returns:
        A list of lists in the format [thread,model] of shape (len(threads), len(models)).        
        For example: out[0] holds threads[0] results on all models, out[1]: threads[1] on all models, ...
    """    

    th_list = []
    for in_text in in_list:
        th = Thread.make_INST_IN(inst_text, in_text)
        th_list.append(th)

    if in_titles is None:
        in_titles = in_list

    out = thread_multigen(th_list,                     
                          model_names=model_names, 
                          text=text,
                          csv=csv,
                          gencall=gencall,
                          genconf=genconf,
                          out_keys=out_keys,
                          thread_titles=in_titles)

    return out







def cycle_gen_print(in_list: list[str],
                    inst_text: str,                                                
                    model_names: list[str],
    
                    gencall: Optional[Callable] = None,                   
                    genconf: Optional[GenConf] = None,
    
                    out_keys: list[str] = ["text","dic", "value"],

                    json_kwargs: dict = {"indent": 2,
                                         "sort_keys": False,
                                         "ensure_ascii": False}
                    ):
    """For a list of models, sequentially grow a Thread with model responses to given IN messages and print the results.

    Works by doing:
    
    1. Generate an INST+IN prompt for a list of models. (Same INST for all).
    2. Append the output of each model to its own Thread.
    3. Append the next IN prompt and generate again. Back to 2.

    Actual generation for each model is implemented by an optional Callable with this signature:
        def gencall(model: Model,
                    thread: Thread,
                    genconf: GenConf) -> GenOut

    Args:
        in_list: List of IN messages to initialize Threads.
        inst_text: The common INST to use in all models.
        model_names: A list of Models names.
        gencall: Callable function that does the actual generation. Defaults to None, which will use a text generation default function.
        genconf: Model generation configuration to use in models. Defaults to None, meaning default values.
        out_keys: A list with GenOut members to output. Defaults to ["text","dic", "value"].
        json_kwargs: JSON dumps() configuration. Defaults to {"indent": 2, "sort_keys": False, "ensure_ascii": False }.
    """

    assert isinstance(model_names, list), "model_names must be a list of strings"

    if gencall is None:
        gencall = _default_gencall_text
            
    
    n_model = len(model_names)
    n_ins = len(in_list)

    for m in range(n_model):
        
        name = model_names[m]
        model = Models.create(name)
        
        print('=' * 80)
        print(f"Model: {name} -> {model.desc}")
        
        th = Thread(inst=inst_text)
        
        for i in range(n_ins):
            in_text = in_list[i]
            print(f"IN: {in_text}")
            
            th += (MsgKind.IN, in_text)

            out = gencall(model, th, genconf)

            out_dict = out.as_dict()

            print("OUT")
            
            for k in out_keys:
                
                if k in out_dict and out_dict[k] is not None:
                    
                    if k != out_keys[0]: # not first
                        print("-" * 20)

                    val = nice_print(k, out_dict[k], json_kwargs)
                    print(val)
                    
            th += (MsgKind.OUT, out.text)

        del model

    




import pytest

from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args
import enum

from datetime import date, time, datetime

from dataclasses import dataclass, is_dataclass, fields, MISSING

from pydantic import BaseModel, Field

from sibila.json_schema import (
    JSchemaConf,
    get_type,
    get_type_list,
    build_root_json_schema,
    get_final_type,
    create_final_instance,
)



def test_get_type_prims():

    # prim_types
    res = get_type(bool)
    assert res == (bool, None, {})

    res = get_type(int)
    assert res == (int, None, {})

    res = get_type(float)
    assert res == (float, None, {})

    res = get_type(str)
    assert res == (str, None, {})

    res = get_type(Annotated[str, "Item desc"])
    assert res == (str, 'Item desc', {})





    
def test_get_type_enums():

    # List enums
    res = get_type([True,False])
    assert res == (bool, None, {"enum_list": [True, False]})

    res = get_type([1,2,3])
    assert res == (int, None, {"enum_list": [1, 2, 3]})

    res = get_type([1.0,2.0,3.0])
    assert res == (float, None, {"enum_list": [1.0, 2.0, 3.0]})

    res = get_type(["good", "neutral", "bad"])
    assert res == (str, None, {"enum_list": ['good', 'neutral', 'bad']})

    # Annotated is not allowed with [type]

    with pytest.raises(TypeError):
        res = get_type([1,2,"3"])

    class Good():
        a=1
    class Bad():
        a=2
    with pytest.raises(TypeError):
        res = get_type([Good, Bad])



    # Literals
    res = get_type(Literal[True,False])
    assert res == (bool, None, {"enum_list": [True, False]})

    res = get_type(Literal[1,2,3])
    assert res == (int, None, {"enum_list": [1, 2, 3]})

    res = get_type(Annotated[Literal[1,2,3], "Desc"])
    assert res == (int, "Desc", {"enum_list": [1, 2, 3]})

    with pytest.raises(TypeError):
        res = get_type(Literal[1,2.0,3])



    # Enum* classes
    class Color(enum.Enum):
        RED = 1
        GREEN = 2
        BLUE = 3    
    res = get_type(Color)
    assert res == (int, None, {"enum_list": [1, 2, 3]})

    class Color(enum.IntEnum):
        RED = 1
        GREEN = 2
        BLUE = 3    
    res = get_type(Color)
    assert res == (int, None, {"enum_list": [1, 2, 3]})

    class Color(enum.StrEnum):
        RED = "1"
        GREEN = "2"
        BLUE = "3"
    res = get_type(Color)
    assert res == (str, None, {"enum_list": ["1", "2", "3"]})

    class Color(enum.Enum):
        RED = True
        GREEN = False
        BLUE = True
    res = get_type(Color)
    assert res == (bool, None, {"enum_list": [True, False]})


    class Color(enum.IntEnum):
        RED = 1
        GREEN = 2
        BLUE = 3    
    res = get_type(Annotated[Color, "Desc"])
    assert res == (int, "Desc", {"enum_list": [1, 2, 3]})


    class Color(enum.Enum):
        RED = "1"
        GREEN = "2"
        BLUE = 3
    with pytest.raises(TypeError):
        res = get_type(Color)




def test_get_type_datetime():

    res = get_type(datetime)
    assert res == (str, None, {'str_format': 'date-time'})

    res = get_type(Annotated[datetime,"desc"])
    assert res == (str, 'desc', {'str_format': 'date-time'})
    

    res = get_type(date)
    assert res == (str, None, {'str_format': 'date'})

    res = get_type(Annotated[date,"desc"])
    assert res == (str, 'desc', {'str_format': 'date'})


    res = get_type(time)
    assert res == (str, None, {'str_format': 'time'})

    res = get_type(Annotated[time,"desc"])
    assert res == (str, 'desc', {'str_format': 'time'})



def test_get_type_dataclass():

    @dataclass
    class Invent:
        """Class for keeping track of an item in inventory."""
        name: str
        unit_price: float
        counts_list: list[int]
        anno_field: Annotated[str, "This is anno"] = "30"
        quantity: int = 0

    res = get_type(Invent)
    assert res == (Invent, None, {})

    res = get_type(Annotated[Invent, "Desc"])
    assert res == (Invent, 'Desc', {})





def test_get_type_BaseModel():

    class UserDetail(BaseModel):
        """ Details of a user """
        name: str
        age: int
    
    res = get_type(UserDetail)
    assert res == (UserDetail, None, {})

    res = get_type(Annotated[UserDetail, "Desc"])
    assert res == (UserDetail, 'Desc', {})









def test_get_type_list():

    # list[prim_type]
    res = get_type_list(list[int])
    assert res == (int, None, {}, None)

    res = get_type_list(Annotated[list[str], "List desc"])
    assert res == (str, None, {}, 'List desc')

    res = get_type_list(list[Annotated[str, "Item desc"]])
    assert res == (str, 'Item desc', {}, None)

    res = get_type_list(Annotated[list[Annotated[str, "Item desc"]], "List desc"])
    assert res == (str, 'Item desc', {}, 'List desc')



    # enums
    res = get_type_list(list[Literal["a", "b"]])
    assert res == (str, None, {"enum_list": ['a', 'b']}, None)

    class Color(enum.IntEnum):
        RED = 1
        GREEN = 2
        BLUE = 3    

    res = get_type_list(list[Color])
    assert res == (int, None, {"enum_list": [1, 2, 3]}, None)



    # datetime
    res = get_type_list(list[datetime])
    assert res == (str, None, {'str_format': 'date-time'}, None)

    res = get_type_list(Annotated[list[datetime], "List desc"])
    assert res == (str, None, {'str_format': 'date-time'}, 'List desc')


    res = get_type_list(list[date])
    assert res == (str, None, {'str_format': 'date'}, None)

    res = get_type_list(Annotated[list[date], "List desc"])
    assert res == (str, None, {'str_format': 'date'}, 'List desc')


    res = get_type_list(list[time])
    assert res == (str, None, {'str_format': 'time'}, None)

    res = get_type_list(Annotated[list[time], "List desc"])
    assert res == (str, None, {'str_format': 'time'}, 'List desc')


    # dataclass
    @dataclass
    class Invent:
        """Class for keeping track of an item in inventory."""
        name: str
        unit_price: float
        counts_list: list[int]
        anno_field: Annotated[str, "This is anno"] = "30"
        quantity: int = 0

    res = get_type_list(list[Invent])
    assert res == (Invent, None, {}, None)

    res = get_type_list(Annotated[list[Invent], "list desc"])
    assert res == (Invent, None, {}, 'list desc')




    # BaseModel
    class UserDetail(BaseModel):
        """ Details of a user """
        name: str
        age: int

    res = get_type_list(list[UserDetail])
    assert res == (UserDetail, None, {}, None)
    
    res = get_type_list(list[Annotated[UserDetail, "Item desc"]])
    assert res == (UserDetail, 'Item desc', {}, None)

    res = get_type_list(Annotated[list[UserDetail], "List desc"])
    assert res == (UserDetail, None, {}, 'List desc')

    res = get_type_list(Annotated[list[Annotated[UserDetail, "Item desc"]], "List desc"])
    assert res == (UserDetail, "Item desc", {}, 'List desc')

    class Good():
        a=1
    with pytest.raises(TypeError):
        res = get_type_list(list[Good])

    with pytest.raises(TypeError):
        res = get_type_list(list[list[int]])








def test_build_root_json_schema():
    """ Only base types, not BaseModel, whose json_schema generation is external """

    res = build_root_json_schema(int, "output")
    assert res == ({'properties': {'output': {'type': 'integer'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)


    res = build_root_json_schema(list[int], "output")
    assert res == ({'properties': {'output': {'items': {'type': 'integer'}, 'type': 'array'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)


    # datetime
    res = build_root_json_schema(datetime, "output")
    assert res == ({'properties': {'output': {'format': 'date-time', 'type': 'string'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)

    res = build_root_json_schema(list[datetime], "output")
    assert res == ({'properties': {'output': {'items': {'format': 'date-time', 'type': 'string'},
                        'type': 'array'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)



    res = build_root_json_schema(date, "output")
    assert res == ({'properties': {'output': {'format': 'date', 'type': 'string'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)

    res = build_root_json_schema(list[date], "output")
    assert res == ({'properties': {'output': {'items': {'format': 'date', 'type': 'string'},
                        'type': 'array'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)



    res = build_root_json_schema(time, "output")
    assert res == ({'properties': {'output': {'format': 'time', 'type': 'string'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)

    res = build_root_json_schema(list[time], "output")
    assert res == ({'properties': {'output': {'items': {'format': 'time', 'type': 'string'},
                    'type': 'array'}},
                    'required': ['output'],
                    'type': 'object'},
                    True)


    # dataclass
    @dataclass
    class Invent:
        """Class for keeping track of an item in inventory."""
        name: str
        counts_list: list[int]
        unit_price: float = 0.1
        anno_field: Annotated[str, "This is anno"] = "30"

    res = build_root_json_schema(Invent, "output")
    assert res == ({'description': 'Class for keeping track of an item in inventory.',
                    'properties': {'name': {'type': 'string'},
                                'counts_list': {'items': {'type': 'integer'}, 'type': 'array'},
                                'unit_price': {'type': 'number', 'default': 0.1},
                                'anno_field': {'description': 'This is anno',
                                    'type': 'string',
                                    'default': '30'}
                                },
                    'required': ['name', 'counts_list'],
                    'type': 'object'},
                    False)
    
    res = build_root_json_schema(Annotated[Invent, "Invent desc"], "output")
    assert res == ({'description': 'Invent desc',
                    'properties': {'name': {'type': 'string'},
                    'counts_list': {'items': {'type': 'integer'}, 'type': 'array'},
                    'unit_price': {'type': 'number', 'default': 0.1},
                    'anno_field': {'description': 'This is anno',
                        'type': 'string',
                        'default': '30'}},
                    'required': ['name', 'counts_list'],
                    'type': 'object'},
                    False)

    res = build_root_json_schema(list[Invent], "output")
    assert res == ({'properties': {'output': {'items': {'description': 'Class for keeping track of an item in inventory.',
                                        'properties': {'name': {'type': 'string'},
                                        'counts_list': {'items': {'type': 'integer'}, 'type': 'array'},
                                        'unit_price': {'type': 'number', 'default': 0.1},
                                        'anno_field': {'description': 'This is anno',
                                        'type': 'string',
                                        'default': '30'}},
                                        'required': ['name', 'counts_list'],
                                        'type': 'object'},
                                        'type': 'array'}
                                    },
                    'required': ['output'],
                    'type': 'object'},
                    True)


    @dataclass
    class Inner:
        """Class Inner."""
        iname: str
        ifield: int = 2
        
    @dataclass
    class Invent:
        """Class for keeping track of an item in inventory."""
        inner: Inner
        num: int = 4

    res = build_root_json_schema(Invent, "output")
    assert res == ({'description': 'Class for keeping track of an item in inventory.',
                    'properties': {'inner': {'description': 'Class Inner.',
                        'properties': {'iname': {'type': 'string'},
                        'ifield': {'type': 'integer', 'default': 2}},
                        'required': ['iname'],
                        'type': 'object'},
                    'num': {'type': 'integer', 'default': 4}},
                    'required': ['inner'],
                    'type': 'object'},
                    False)
    














def test_get_final_type():

    res = get_final_type(int)
    assert res == (int, False)

    res = get_final_type([1.0,2.0])
    assert res == (float, False)

    res = get_final_type(Literal[1,2])
    assert res == (int, False)

    class Color(enum.IntEnum):
        RED = 1
        GREEN = 2
        BLUE = 3    
    res = get_final_type(Color)
    assert res == (Color, False)


    # datetime
    res = get_final_type(datetime)
    assert res == (datetime, False)

    res = get_final_type(date)
    assert res == (date, False)

    res = get_final_type(time)
    assert res == (time, False)


    # BaseModel
    class UserDetail(BaseModel):
        """ Details of a user """
        name: str
        age: int

    res = get_final_type(UserDetail)
    assert res == (UserDetail, False)


    # lists
    target = list[UserDetail]
    res = get_final_type(target)
    assert res == (UserDetail, True)

    target = Annotated[list[UserDetail], "Desc"]
    res = get_final_type(target)
    assert res == (UserDetail, True)





def test_create_final_instance():

    res = create_final_instance(datetime, False, "1984-05-11T15:10:00", JSchemaConf())
    assert res == datetime(1984, 5, 11, 15, 10)

    res = create_final_instance(datetime, True, 
                                ["1984-05-11T15:10:00", "2014-06-05T00:00:00"], 
                                JSchemaConf())
    assert res == [datetime(1984, 5, 11, 15, 10), datetime(2014, 6, 5, 0, 0)]


    res = create_final_instance(date, False, "1984-05-11", JSchemaConf())
    assert res == date(1984, 5, 11)

    res = create_final_instance(date, True, 
                                ["1984-05-11", "2014-06-05"], 
                                JSchemaConf())
    assert res == [date(1984, 5, 11), date(2014, 6, 5)]


    res = create_final_instance(time, False, "15:10:00", JSchemaConf())
    assert res == time(15, 10)

    res = create_final_instance(time, True, 
                                ["15:10:00", "00:00:00"], 
                                JSchemaConf())
    assert res == [time(15, 10), time(0, 0)]






"""JSON Schema generator functions"""

from typing import Any, Optional, Union, Literal, Annotated, get_origin, get_args
from enum import Enum

from copy import copy

import json

import logging
logger = logging.getLogger(__name__)

from datetime import date, time, datetime

from dataclasses import dataclass, is_dataclass, field, fields, MISSING

from pydantic import BaseModel, TypeAdapter, ValidationError

from .utils import (
    is_subclass_of,
    synth_desc
)



@dataclass
class JSchemaConf:
    """Configuration for JSON schema massaging and validation."""

    # json_schema_massage() configuration:
    resolve_refs: bool = True
    """Set for $ref references to be resolved and replaced with actual definition."""
    
    collapse_single_combines: bool = True
    """Any single-valued "oneOf"/"anyOf" is replaced with the actual value."""
    
    description_from_title: int = 0
    """If a value doesn't have a description entry, make one from its title or name.

    - 0: don't make description from name
    - 1: copy title or name to description
    - 2: 1: + capitalize first letter and convert _ to space: class_label -> "class label". """
    
    force_all_required: bool = False
    """Force all entries in an object to be required (except removed defaults if remove_with_default=True)."""
    
    remove_with_default: bool = False
    """Delete any values that have a "default" annotation."""

    default_to_last: bool = True
    """Move any default value entry into the last position of properties dict."""

    additional_allowed_root_keys: list = field(default_factory=list)
    """By default only the following properties are allowed in schema's root:
         description, properties, type, required, additionalProperties, allOf, anyOf, oneOf, not
       Add to this list to allow additional root properties."""

    # ideas for more simplification:
    #   force_no_additional_properties: sets "additionalProperties": false on all type=object

    
    # pydantic_obj_from_json() configuration:
    pydantic_strict_validation: Optional[bool] = None
    """Validate JSON values in a strict manner or not. None means validate individually per each value in the obj. (for example in pydantic with: Field(strict=True))."""

    
    def clone(self):
        """Return a copy of this configuration."""
        return copy(self)







# after https://github.com/pydantic/pydantic/issues/889#issuecomment-850312496
def json_schema_resolve_refs(schema: dict,
                             del_root_defs: bool = True,
                             max_tries: int = 100) -> dict:
    
    schema = schema.copy()

    root_defs = set()
    
    def replace_value_in_dict(item, original_schema):
        if isinstance(item, list):
            return [replace_value_in_dict(i, original_schema) for i in item]
            
        elif isinstance(item, dict):
            if list(item.keys()) == ['$ref']:
                definitions = item['$ref'][2:].split('/')
                root_defs.add(definitions[0])
                res = original_schema.copy()
                for definition in definitions:
                    res = res[definition]
                return res
                
            else:
                return {key: replace_value_in_dict(i, original_schema) for key, i in item.items()}
                
        else:
            return item        
    
    
    for i in range(max_tries):
        if '$ref' not in json.dumps(schema):
            break
        schema = replace_value_in_dict(schema.copy(), schema.copy())

    # remove root defs
    if del_root_defs:
        for k in root_defs:
            if k in schema.keys():
                del schema[k]
    
    return schema




def json_schema_massage(sch: dict,
                        schemaconf: JSchemaConf = JSchemaConf(),
                        DEB: bool = False) -> dict:

    """
    Massages JSON schema object to simplify as much as possible and remove all non-essential keys.
    Resolves $refs and eliminates definitions.
    """
   
    # resolve $refs -> defs
    if schemaconf.resolve_refs:
        sch = json_schema_resolve_refs(sch)
        if DEB:
            from pprint import pprint
            print("resolved:")
            pprint(sch)
            
   
    # root: clean all other than:
    allowed_in_root = ["description",
                       "properties", 
                       "type", 
                       "required",
                       "additionalProperties",
                       "allOf", "anyOf", "oneOf", "not"] + schemaconf.additional_allowed_root_keys
        
    root = {
        k: v for k, v in sch.items() if k in allowed_in_root
    }
    

    
    def recurse_object_or_items(dic: dict):
        if "title" in dic:
            if schemaconf.description_from_title and "description" not in dic:
                dic["description"] = synth_desc(schemaconf.description_from_title - 1, 
                                                dic["title"])

            del dic["title"]

        clean(dic)

        
    def recurse_combine(lis: list):
        for dic in lis:
            if "title" in dic:
                if schemaconf.description_from_title and "description" not in dic:
                    dic["description"] = synth_desc(schemaconf.description_from_title - 1, 
                                                    dic["title"])
    
                del dic["title"]

            clean(dic)

    
    def clean(root: dict):

        if DEB: print("===", root.keys())
        
        if "properties" in root:
            required_keys = []

            vars_dict = root["properties"]

            # for each prop var
            kl = list(vars_dict.keys())
            if DEB: print("---properties", kl)

            # for dict fields in properties
            for k in kl:

                if schemaconf.force_all_required:
                    required_keys.append(k)
                
                default_to_last_value = None
                if "default" in vars_dict[k]:
                    if schemaconf.remove_with_default: # skip entries with default values
                        del vars_dict[k]
                        continue
                    elif schemaconf.default_to_last: # save and delete
                        default_to_last_value = [vars_dict[k]["default"]] # because default can be None
                        del vars_dict[k]["default"]
        
                recurse_object_or_items(vars_dict[k])

                if default_to_last_value is not None:
                    vars_dict[k]["default"] = default_to_last_value[0]
            
            if schemaconf.force_all_required:
                root["required"] = required_keys

        
        elif "items" in root:            
            vars_dict = root["items"]

            if DEB: print("---items", list(vars_dict.keys()))
            
            recurse_object_or_items(vars_dict)

        
        if "allOf" in root:
            which = "allOf"
        elif "anyOf" in root:
            which = "anyOf"
        elif "oneOf" in root:
            which = "oneOf"
        else:
            which = None

        if which is not None: # a combine
            if DEB: print(f"---{which}")                

            if len(root[which]) == 1 and schemaconf.collapse_single_combines:
                item = root[which][0]
                del root[which]
                root.update(item)
                recurse_object_or_items(root)
            else:
                recurse_combine(root[which])

                
    clean(root)

    return root




# ============================================================================ Pydantic BaseModel

def json_schema_from_pydantic(cls: BaseModel) -> dict:
    return cls.model_json_schema()


def pydantic_obj_from_json(cls: BaseModel, 
                           obj_init: dict,
                           schemaconf: Optional[JSchemaConf] = None
                           ) -> BaseModel:

    """
    For info on strict_validation see:
    https://docs.pydantic.dev/latest/concepts/strict_mode/

    strict_validation can be applied per class:
        class MyModel(BaseModel):
            model_config = ConfigDict(strict=False)
            
    or per Field:
        class Model(BaseModel):
            x: int = Field(strict=True)
            y: int = Field(strict=False)                

    https://docs.pydantic.dev/latest/api/type_adapter/
    """
    
    if schemaconf is None:
        schemaconf = JSchemaConf()

    adapter = TypeAdapter(cls)
    try:
        obj = adapter.validate_python(obj_init, 
                                      strict=schemaconf.pydantic_strict_validation)
        return obj
        
    except ValidationError as e:
        raise TypeError(str(e))
            



    


# ============================================================================ JSON Schema construction from types

def is_prim_type(type_: Any):
    
    # cannot use issubclass which would accept IntEnum as int
    return (type_ is str or 
            type_ is float or 
            type_ is int or 
            type_ is bool)





def get_type(type_: Any) -> tuple:
    """Identifies a simple type, that can be generated by model.
    Argument type_ can be one of the following non-list types:

    - prim_type:

        - bool
        - int
        - float
        - str
        
    - enums:

        - [1, 2, 3] or ["a","b"] - all items of the same prim_type
        - Literal['year', 'name'] - all items of the same prim_type
        - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

    - datetime/date/time

    - a list in the form:
        - list[type]
        
        For example list[int]. The list can be annotated:
            Annotated[list[T], "List desc"]
        And/or the list item type can be annotated:
            list[Annotated[T, "Item desc"]]

    - dataclass with fields of the above supported types (or dataclass).

    - Pydantic BaseModel

    All types can be Annotated[T, "Desc"], for example: 
        count: int
    Can be annotated as:
        count: Annotated[int, "How many units?"]

    Args:
        type_: Type to be simplified, one of the above.

    Returns:
        Returns (type_, anno_desc, options) or (None,,) if not a supported type.
        options: optional keys
        "str_format": str
        "enum_list": []
    """

    anno_desc = None
    options = {}

    if get_origin(type_) is Annotated:
        args = list(get_args(type_))
        type_ = args[0]
        if len(args) > 1:
            anno_desc = args[1]


    if (is_prim_type(type_) or
        is_dataclass(type_) or
        is_subclass_of(type_, BaseModel)):
        ...

    elif is_subclass_of(type_, datetime):
        type_ = str
        options["str_format"] = "date-time"

    elif is_subclass_of(type_, date):
        type_ = str
        options["str_format"] = "date"

    elif is_subclass_of(type_, time):
        type_ = str
        options["str_format"] = "time"

    else: # enums
        type_, enum_list = get_enum_type(type_)

        if type_ is not None:
            options["enum_list"] = enum_list

    return type_, anno_desc, options
    



def get_enum_type(type_: Any) -> tuple:
    """Identifies an enumeration that can be generated by model. 

    Args:
        type_: Can be one of the following forms:

            - [1, 2, 3] or ["a","b"] - all items of the same prim_type
            - Literal['year', 'name'] - all items of the same prim_type
            - Enum, EnumInt, EnumStr, (Enum, int),... - all items of the same prim_type

    Raises:
        TypeError: If not a valid enumeration.

    Returns:
        A tuple in the form enum_type, list_of_enum_values.
        Where enum_type if a value accepted by get_type().
    """

    enum_list = None

    if get_origin(type_) is Literal:
        enum_list = list(get_args(type_))
        type_ = type(enum_list[0])

    elif is_subclass_of(type_, Enum):
        enum_list = [e.value for e in type_]
        type_ = type(enum_list[0])

    elif isinstance(type_, list): # enum as value list: ["a","b"]
        enum_list = type_[:]
        type_ = type(type_[0])

    else:
        type_ = None

    if type_ is not None: # enum consistency checks
        if not is_prim_type(type_):
            raise TypeError(f"Base type is not one of bool, int, float, str: '{type_}'")

        if not enum_list:
            raise TypeError("Enum must have at least one value")

        if not all([type_ is type(e) for e in enum_list]):
            raise TypeError(f"All enum values must have the same type in '{enum_list}'")
        
    return type_, enum_list




def get_type_list(type_: Any) -> tuple:
    """Identify a type which is a list of types accepted by get_type().

    Args:
        type_: A type in the form: 
            - list[type]

                For example list[int]. The list can be annotated:
                    Annotated[list[T], "List desc"]
                And/or the list item type can be annotated:
                    list[Annotated[T, "Item desc"]]

    Raises:
        TypeError: If not a valid list.

    Returns:
        A tuple in the form: item_type, item_anno_desc, item_options, list_anno_desc.
    """

    item_anno_desc = None
    item_options = {}
    list_anno_desc = None

    if get_origin(type_) is Annotated:
        args = list(get_args(type_))
        type_ = args[0]
        if len(args) > 1:
            list_anno_desc = args[1]

    if get_origin(type_) is list: # list[type]
        args = list(get_args(type_))
        type_ = args[0]

    else:
        type_ = None

    if type_ is not None:
        type_, item_anno_desc, item_options = get_type(type_)

        if type_ is None: # list type consistency
            raise TypeError(f"List item type is not a supported type like bool, int, float, str, enum, dataclass, datetime, BaseModel: '{type_}'")

    return type_, item_anno_desc, item_options, list_anno_desc

    





def build_type_json_schema(type_: Any, 
                           desc: Optional[str] = None,
                           options: dict = {},
                           default: Optional[Any] = None) -> dict:
    """Render a JSON Schema specification for an accepted simple type.

    Args:
        type_: A type supported by get_type().
        desc: Optional description for field. Defaults to None, but if given, will override any Annotated type's description.
        default: A default value for field. Defaults to {}.
        format: A default value for strings. Defaults to None.

    Returns:
        A dict whose json.dumps() serialization is a valid JSON Schema specification for the given type_ arg.
    """

    out_type, anno_desc, options2 = get_type(type_)
    if out_type is None:
        raise TypeError(f"Unsupported type: '{type(type_)}'")
    
    if desc is None:
        desc = anno_desc

    if is_dataclass(out_type):
        out = build_dataclass_object_json_schema(out_type)

        if desc is not None:
            out["description"] = desc

    elif is_subclass_of(out_type, BaseModel):
        out = out_type.model_json_schema()

        if desc is not None:
            out["description"] = desc

    else: # prim_type or enum
        out = {}

        if desc:
            out["description"] = desc

        enum_list = options.get("enum_list")
        if enum_list is None:
            enum_list = options2.get("enum_list")
        if enum_list is not None:
            out["enum"] = enum_list

        str_format = options.get("str_format")
        if str_format is None:
            str_format = options2.get("str_format")
        if str_format is not None:
            if out_type is not str:
                raise TypeError("Arg str_format is only valid for str type.")

            out["format"] = str_format

        out["type"] = get_json_type(out_type)

        if default is not None:
            if not isinstance(default, out_type):
                raise TypeError(f"Arg default is not of type {out_type}.")

            out["default"] = default

    return out



def build_array_json_schema(item_type: Any,
                            item_desc: Optional[str] = None,
                            item_options: dict = {},
                            list_desc: Optional[str] = None) -> dict:
    """Render a JSON Schema specification for an list/array.

    Args:
        item_type: Type for array items.
        item_desc: Optional description for an array item. Defaults to None, but if given, will override any Annotated type's description.
        item_options: Optional array item options. Defaults to {}.
        list_desc: Optional description for the entire array. Defaults to None.

    Returns:
        A dict whose json.dumps() serialization is a valid JSON Schema specification for the given array.
    """

    items_repr = build_type_json_schema(item_type, 
                                        item_desc,
                                        options=item_options)

    out: dict[str,Any] = {}

    if list_desc:
        out["description"] = list_desc

    out["items"] = items_repr

    out["type"] = "array"

    return out








def build_type_or_array_json_schema(type_: Any,
                                    default: Optional[Any] = None) -> tuple[dict,bool]:
    """Build a JSON schema for given simple type or array. See build_*() functions for details."""

    # type list
    item_type, item_desc, item_options, list_desc = get_type_list(type_)    
    if item_type is not None:            
        # build json schema for list of type_
        schema = build_array_json_schema(item_type, 
                                         item_desc, 
                                         item_options,
                                         list_desc)
        is_object = False

    else: # prim, enum, datetime, dataclass, BaseModel
        out_type, desc, options = get_type(type_)
        if out_type is None:
            raise TypeError(f"Unknown target type '{type_}'")

        schema = build_type_json_schema(out_type, 
                                        desc,
                                        options=options,
                                        default=default)
        
        is_object = schema.get("type") == "object"

    return schema, is_object










def build_object_json_schema(properties_repr: dict,
                             desc: Optional[str] = None,
                             required_keys: Optional[list[str]] = None) -> dict:
    """Build a JSON schema for an object type."""

    out: dict[str,Any] = {}

    if desc:
        out["description"] = desc

    out["properties"] = properties_repr

    all_keys = list(properties_repr.keys())
    if required_keys is None:
        required_keys = all_keys
    else:
        if not all([k in all_keys for k in required_keys]):
            raise ValueError("Arg required_keys has unknown properties keys")

    out["required"] = required_keys        

    out["type"] = "object"

    return out





def build_dataclass_object_json_schema(type_: Any) -> dict:
    """Build a JSON schema for a dataclass type."""

    desc = type_.__doc__

    props = {}
    required = []

    flds = fields(type_)
    for fl in flds:
        name = fl.name
        prop_type = fl.type # can be Annotated[] -> desc
        if fl.default is MISSING:
            default = None
            required.append(name)
        else:
            default = fl.default

        prop_schema, _ = build_type_or_array_json_schema(prop_type,
                                                         default=default)

        props[name] = prop_schema

    if not required: # we can't have an empty dataclass: make all required
        required = list(props.keys())

    out = build_object_json_schema(props,
                                   desc,
                                   required)

    return out






def build_root_json_schema(type_: Any, 
                           output_key_name: str) -> tuple[dict,bool]:
    """Build a JSON schema for a type: simple type, list, array, dataclass or Pydantic BaseModel."""

    schema, is_object = build_type_or_array_json_schema(type_)

    if not is_object: # need to create an object in root
        schema = build_object_json_schema({output_key_name: schema})

    return schema, not is_object








def get_final_type(type_: Any) -> Any:
    """Final type is the actual value type that will be instantiated with the model's response."""

    is_list = False

    # dig through list and annotations until a valid type is found
    while True:
        orig = get_origin(type_)
        if orig in (Annotated, list): # get rid of annotations and list[]
            is_list = is_list or orig == list

            args = list(get_args(type_))
            type_ = args[0]
            continue


        if (is_dataclass(type_) or
            is_subclass_of(type_, BaseModel) or
            is_subclass_of(type_, datetime) or
            is_subclass_of(type_, date) or
            is_subclass_of(type_, time)):
            break
        
        elif isinstance(type_, list): # enum list: ["a", "b"]
            type_ = type(type_[0])
            break
        
        elif get_origin(type_) is Literal: # Literals resolve to their prim_types
            enum_list = list(get_args(type_))
            type_ = type(enum_list[0])
            break

        elif is_subclass_of(type_, Enum): # Enum types resolve to themselves
            break

        elif (type_ is str or # prim_types
              type_ is float or 
              type_ is int or 
              type_ is bool):
            break
        
        else:
            raise TypeError(f"Unknown final type: '{type_}'")

    return type_, is_list





def create_final_instance(type_: Any,
                          is_list: bool,
                          val: Any,
                          schemaconf: Optional[JSchemaConf] = None) -> Any:
    
    """Instantiate an object with the model's answer."""
    
    if schemaconf is None:
        schemaconf = JSchemaConf()

    def create_item(type_: Any,
                    val: Any) -> Any:

        if is_dataclass(type_):
            if not isinstance(val, dict):
                raise TypeError(f"Expecting dict to initialize dataclass but got: '{val}'")
            obj = type_(**val)
            return obj

        elif is_subclass_of(type_, BaseModel):
            if not isinstance(val, dict):
                raise TypeError(f"Expecting dict to initialize BaseModel but got: '{val}'")
            obj = pydantic_obj_from_json(type_, 
                                         val,
                                         schemaconf=schemaconf)
            return obj
        
        elif is_subclass_of(type_, datetime):
            if not isinstance(val, str):
                raise TypeError(f"Expecting str to initialize datetime but got: '{val}'")
            return datetime.fromisoformat(val)

        elif is_subclass_of(type_, date):
            if not isinstance(val, str):
                raise TypeError(f"Expecting str to initialize date but got: '{val}'")
            return date.fromisoformat(val)

        elif is_subclass_of(type_, time):
            if not isinstance(val, str):
                raise TypeError(f"Expecting str to initialize time but got: '{val}'")
            return time.fromisoformat(val)

        elif is_subclass_of(type_, Enum):
            return type_(val)
        
        # Literals also resolve to their prim_types        
        elif (type_ is str or 
              type_ is float or 
              type_ is int or 
              type_ is bool):
            value = type_(val) 
            return value
        
        else:
            raise TypeError(f"Unexpected value of type '{type_}' for value '{val}'")


    if is_list:
        if type(val) is not list: # check, just in case...
            raise TypeError(f"Value is not a list: '{val}'")
        
        out = []
        for item_val in val:
            out.append(create_item(type_, item_val))

        return out

    else:
        return create_item(type_, val)








def get_json_type(t: Any) -> str:
    """Translate from native types to JSON's."""
    JSON_TYPE_FROM_PY_TYPE = {
        str: "string",
        float: "number",            
        int: "integer",            
        bool: "boolean"
    }
    if t not in JSON_TYPE_FROM_PY_TYPE:
        raise TypeError(f"Unknown type '{t}'")
    
    return JSON_TYPE_FROM_PY_TYPE[t]



# Useful:
# string formats: https://json-schema.org/understanding-json-schema/reference/string#built-in-formats
# Online JSON schema validator: https://www.jsonschemavalidator.net/
# JSON pretty formatter: https://jsonformatter.org/json-pretty-print



"""Text chunk splitter for RAG-like applications."""


from typing import Any, Optional, Union, Callable
from copy import copy
import re


class RecursiveTextSplitter:
    """
    When using a token_len len_fn, returned chunks might be sightly smaller than chunk size, if one of the separators is space, because tokens may already assume space as a prefix.
    """

    def __init__(self,
                 chunk_size: int,
                 chunk_overlap: int = 0,
                 seps: Optional[list[str]] = None,
                 len_fn: Optional[Callable[[str], int]] = None
                 ):
        """
        chunk_size and chunk_overlap are in whatever units len_fn returns (chars or tokens for example).
        """
        
        if chunk_overlap >= (chunk_size // 2):
            raise ValueError(f"chunk_overlap ({chunk_overlap}) must be less than half the chunk_size ({chunk_size})")

        if chunk_size < 3:
            raise ValueError("chunk_size must be at least 3")
        
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.len_fn = len_fn or len
        
        self.seps = seps or ["\n\n", "\n", " ", ""]
        if self.seps[-1] != "":
            self.seps.append("")
        


    def split(self,
              text: Optional[Union[str,list[str]]] = None,
              path: Optional[Union[str,list[str]]] = None,
              seps: Optional[list[str]] = None,
              len_fn: Optional[Callable[[str], int]] = None
              ) -> list[str]:

        if (text is not None) + (path is not None) != 1:
            raise ValueError("Only one of text or path can be given")
        
        if path is not None:
            if isinstance(path, str):
                path = [path]
            text = []
            for p in path:
                with open(p, "r", encoding="utf-8") as f:
                    text.append(f.read()) 
        elif isinstance(text, str):
            text = [text]
                
        out = []

        for t in text:  # type: ignore[union-attr]
            out += self._split(t, seps, len_fn)
    
        return out
    


    

    def _split(self,
               text: str,
               seps: Optional[list[str]] = None,
               len_fn: Optional[Callable[[str], int]] = None
               ) -> list[str]:
        # print(text, out)
        
        seps = seps or self.seps
        assert seps and seps[-1] == ""

        sep = seps[0]
        if sep != "":
            spl = text.split(sep)
        else:
            spl = list(text)

        len_fn = len_fn or self.len_fn
        sep_len = len_fn(sep)
        
        size, over = self.chunk_size, self.chunk_overlap

        out = []
        accum = "" # only accumulates up to size
        while spl:
            head = spl[0]
            if head: # avoid empty
           
                if len_fn(accum) + sep_len + len_fn(head) <= size: 
                    # fits in accum: merge
                    if accum:
                        accum += sep
                    accum += head
                    
                elif accum and len_fn(head) <= (size - over + sep_len): 
                    # can emit accum and set accum=overlap+head
                    out.append(accum)
                    if over: # get chunk overlap from accum
                        # backtrack along sep till len() >= over size
                        if sep != "":
                            over_text = ""
                            ar = accum.split(sep)
                            for r in ar[-1::-1]:
                                if len_fn(r) + (sep_len * int(bool(over_text))) + len_fn(over_text) >= over:
                                    break
                                else:
                                    if over_text:
                                        over_text = sep + over_text
                                    over_text = r + over_text
                        else:
                            over_text = accum[-over:]
                        
                        head = over_text + sep + head
                    
                    accum = head
                    
                else: # head doesn't fit: recurse into next separator
                    if accum: # emit accum if any
                        out.append(accum)
                        accum = ""
                        
                    out += self._split(head, seps[1:])
            
            spl = spl[1:]
            
        if accum: # emit accum if any
            out.append(accum)

        return out

            
            



"""Model generation configurations and results.

- GenConf: Model generation configuration, used in Model.gen() and variants.
- GenRes: Generaion result enum.
- GenError: Generation exception.
- GenOut: Results of model generation.
"""

from typing import Any, Optional, Union, Callable
from typing_extensions import Self
from dataclasses import dataclass, field, asdict
from enum import IntEnum

import json
from copy import copy

import logging
logger = logging.getLogger(__name__)




@dataclass
class GenConf:
    """Model generation configuration, used in Model.gen() and variants."""
    
    max_tokens: int = 0 
    """Max generated token length. 0 means all available up to output context size (which equals: model.ctx_len - in_prompt_len)"""
    
    stop: Union[str, list[str]] = field(default_factory=list)
    """List of generation stop text sequences"""
    
    temperature: float = 0.
    """Generation temperature. Use 0 to always pick the most probable output, without random sampling. Larger positive values will produce more random outputs."""

    top_p: float = 0.9
    """Nucleus sampling top_p value. Only applies if temperature > 0."""

    format: str = "text"
    """Output format: "text" or "json". For JSON output, text is validated as in json.loads().
    Thread msgs must explicitly request JSON output or a warning will be emitted if string json not present
    (this is automatically done in Model.json() and related calls).
    """

    json_schema: Union[str,dict,None] = None
    """A JSON schema to validate the JSON output.
    Thread msgs must list the JSON schema and request its use; must also set the format to "json".
    """
    
    
    def __call__(self,
                 **kwargs: Any) -> Self:
        """Return a copy of the current GenConf updated with values in kwargs. Doesn't modify object.

        Args:
            **kwargs: update settings of the same names in the returned copy.

        Raises:
            KeyError: If key does not exist.

        Returns:
            A copy of the current object with kwargs values updated. Doesn't modify object.
        """

        ret = copy(self)

        for k,v in kwargs.items():
            if not hasattr(ret, k):
                raise KeyError(f"No such key '{k}'")
            setattr(ret, k,v)

        return ret


    def clone(self) -> Self:
        """Return a copy of this configuration."""
        return copy(self)
        
    def as_dict(self) -> dict:
        """Return GenConf as a dict."""
        return asdict(self)

    @staticmethod
    def from_dict(dic: dict) -> Any: # Any = GenConf
        return GenConf(**dic)




class GenRes(IntEnum):
    """Model generation result."""

    OK_STOP = 1 
    """Generation complete without errors."""

    OK_LENGTH = 0
    """Generation stopped due to reaching max_tokens."""

    ERROR_JSON = -1 
    """Invalid JSON: this is often due to the model returning OK_LENGTH (finished due to max_tokens reached), which cuts off the JSON text."""

    ERROR_JSON_SCHEMA_VAL = -2
    """Failed JSON schema validation."""

    ERROR_JSON_SCHEMA_ERROR = -2
    """JSON schema itself is not valid."""

    ERROR_MODEL = -3
    """Other model internal error."""


    @staticmethod
    def from_finish_reason(finish: str) -> Any: # Any=GenRes
        """Convert a ChatCompletion finish result into a GenRes.

        Args:
            finish: ChatCompletion finish result.

        Returns:
            A GenRes result.
        """
        if finish == 'stop':
            return GenRes.OK_STOP
        elif finish == 'length':
            return GenRes.OK_LENGTH
        elif finish == '!json':
            return GenRes.ERROR_JSON
        elif finish == '!json_schema_val':
            return GenRes.ERROR_JSON_SCHEMA_VAL
        elif finish == '!json_schema_error':
            return GenRes.ERROR_JSON_SCHEMA_ERROR
        else:
            return GenRes.ERROR_MODEL
           
    @staticmethod
    def as_text(res: Any) -> str: # Any=GenRes
        """Returns a friendlier description of the result.

        Args:
            res: Model output result.

        Raises:
            ValueError: If unknown GenRes.

        Returns:
            A friendlier description of the GenRes.
        """

        if res == GenRes.OK_STOP:
            return "Stop"
        elif res == GenRes.OK_LENGTH:
            return "Length (output cut)"
        elif res == GenRes.ERROR_JSON:
            return "JSON decoding error"
            
        elif res == GenRes.ERROR_JSON_SCHEMA_VAL:
            return "JSON SCHEMA validation error"
        elif res == GenRes.ERROR_JSON_SCHEMA_ERROR:
            return "Error in JSON SCHEMA"
            
        elif res == GenRes.ERROR_MODEL:
            return "Model internal error"
        else:
            raise ValueError("Bad/unknow GenRes")



@dataclass
class GenOut:
    """Model output, returned by gen_extract(), gen_json() and other model calls that don't raise exceptions."""
    
    res: GenRes
    """Result of model generation."""
    
    text: str
    """Text generated by model."""
    
    dic: Union[dict,None] = None
    """Python dictionary, output by the structured calls like gen_json()."""

    value: Union[Any, None] = None # Any = accepted type instances, dataclass or Pydantic BaseModel object
    """Initialized instance value, dataclass or Pydantic BaseModel object, as returned in calls like extract()."""
    
    
    def as_dict(self):
        """Return GenOut as a dict."""
        return asdict(self)

    def __str__(self):
        out = f"Error={self.res.as_text(self.res)} text=█{self.text}█"
        if self.dic is not None:
            out += f" dic={self.dic}"
        if self.value is not None:
            out += f" value={self.value}"
        return out




class GenError(RuntimeError, GenOut):
    """Model generation exception, raised when the model was unable to return a response."""

    def __init__(self, 
                 out: GenOut):
        """An error has happened during model generation.

        Args:
            out: Model output
        """

        assert out.res != GenRes.OK_STOP, "OK_STOP is not an error"      

        super().__init__()

        self.res = out.res
        self.text = out.text
        self.dic = out.dic
        self.value = out.value


    @staticmethod
    def raise_if_error(out: GenOut,
                       ok_length_is_error: bool):
        """Raise an exception if the model returned an error

        Args:
            out: Model returned info.
            ok_length_is_error: Should a result of GenRes.OK_LENGTH be considered an error?

        Raises:
            GenError: If an error was returned by model.
        """
        
        if out.res != GenRes.OK_STOP:
            if out.res == GenRes.OK_LENGTH and not ok_length_is_error:
                return # OK_LENGTH to not be considered an error

            raise GenError(out)
            
        
    def __str__(self):
        out = f"Error={self.res.as_text(self.res)} text=█{self.text}█"
        if self.dic is not None:
            out += f" dic={self.dic}"
        if self.value is not None:
            out += f" value={self.value}"
        return out







"""OpenAI remote model access.

- OpenAIModel: Use OpenAI API to access their models.
- OpenAITokenizer: Tokenizer for OpenAI models.
"""


from typing import Any, Optional, Union
import sys, json
from copy import copy

import logging
logger = logging.getLogger(__name__)


from .gen import (
    GenConf,
    GenOut
)

from .thread import (
    Thread,
    MsgKind
)

from .model import (
    MessagesModel,
    Tokenizer
)

from .json_schema import JSchemaConf

try:
    import openai
    has_openai = True
except ImportError:
    has_openai = False
    
try:
    import tiktoken
    has_tiktoken = True
except ImportError:
    has_tiktoken = False



"""
Locate model and token counting data.
More recent models inside each family (GPT-4, GPT-3.5) first
Str values are links to a versioned model.
https://platform.openai.com/docs/models
"""
KNOWN_MODELS: dict = { # name: link | (ctx_len, n_msg_tokens, n_name_tokens)

    # ------------------------------ GPT 4
    "gpt-4-0613":     (8192, 3, 1),
    "gpt-4-32k-0613": (32768, 3, 1),
    "gpt-4-0314":     (8192, 3, 1),
    "gpt-4-32k-0314": (32768, 3, 1),
    
    "gpt-4-1106-preview": (4096, 3, 1), # aka gpt-4-turbo max input: 128000
    
    "gpt-4": "gpt-4-0613",
    "gpt-4-32k": "gpt-4-32k-0613",

    # ------------------------------ GPT 3.5
    "gpt-3.5-turbo-1106": (4096, 3, 1), # max input: 16385
    "gpt-3.5-turbo-0613": (4096, 3, 1),
    "gpt-3.5-turbo-16k-0613": (16385, 3, 1),
    "gpt-3.5-turbo-0301":  (4096, 4, -1),

    "gpt-3.5-turbo": "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k": "gpt-3.5-turbo-16k-0613",

    "gpt-3": "gpt-3.5-turbo-1106",
    "gpt-3.5": "gpt-3.5-turbo-1106",

}

def resolve_model(name: str, 
                  unknown_name_mask: int) -> tuple[str, int, int, int]:
    """Resolve a name string into an existing model from KNOWN_MODELS.
     Resolution depends on unknown_name_mask and will keep removing letters 
     from the end of name and searching existing entries in KNOWN_MODELS.
     
    Args:
        name: The model name to resolve.
        unknown_name_mask: How to deal with unmatched names in name resolution, a mask of:

            - 2: Raise NameError if exact name not found (no generics)
            - 1: Only allow versioned names - raise NameError if generic non-versioned model name used
            - 0: Accept any name that can be resolved from KNOWN_MODELS

    Raises:
        NameError: If not found, according to unknown_name_mask.

    Returns:
        Existing model name from KNOWN_MODELS.
    """
    
    if name not in KNOWN_MODELS: # not an exact match
        
        if unknown_name_mask & 2:
            raise NameError(f"Unknown model '{name}'. Please provide a correct model name. "
                            "Call OpenAIModel.known_models() for a list of supported model names.")
        
        # 1) search first entry that starts with name
        nam = name
        found = None        

        while nam:
            for k in KNOWN_MODELS:
                if k.startswith(nam):
                    found = k
                    break
            if found is not None:
                break

            nam = nam[:-1] # remove last letter and try again
            
        if found is None: # 2) pick first entry in list
            raise NameError(f"Unknown model '{name}'. Please provide an existing model name. "
                            "Call OpenAIModel.known_models() for a list of supported model names.")

        logger.warn(f"Exact match '{name}' not found: using '{found}' model")
        name = found
            
    v = KNOWN_MODELS[name]
    if isinstance(v, str):
        
        if unknown_name_mask & 1:
            raise NameError(f"Unknown model '{name}' (would use '{v}' but unknown_name_mask doesn't allow). "
                            "Please provide a versioned model name or use a more permissive unknown_name_mask. "
                            "Call OpenAIModel.known_models() for a list of supported model names.")
            
        logger.warn(f"'{name}' is a generic name that may be updated over time: resolved as '{v}' model")
        name = v

    t = KNOWN_MODELS[name]
    return name, t[0], t[1], t[2]







class OpenAIModel(MessagesModel):
    """Access an OpenAI model.

    Supports constrained JSON output, via the OpenAI API tools mechanism.
    Ref: https://platform.openai.com/docs/api-reference/chat/create

    Attributes:
        ctx_len: Maximum context length, shared for input + output.
        desc: Model information.
    """

    TOOLS_TOKEN_LEN_FACTOR: float
    """Multiplication factor to use for tools section token length estimation."""

    DEFAULT_TOOLS_TOKEN_LEN_FACTOR: float = 1.2

    def __init__(self,
                 name: str,
                 unknown_name_mask: int = 2,
                 *,
                 
                 # common base model args
                 genconf: Optional[GenConf] = None,
                 schemaconf: Optional[JSchemaConf] = None,
                 tokenizer: Optional[Tokenizer] = None,
                 ctx_len: int = 0,
                
                 # most important OpenAI-specific args
                 api_key: Optional[str] = None,
                 base_url: Optional[str] = None,
                 
                 # OpenAI-specific args
                 openai_init_kwargs: dict = {},
                 ):
        """Create an OpenAI remote model.
        Name resolution depends on unknown_name_mask and will keep removing letters 
        from the end of name and searching existing entries in penAIModel.known_models().

        Args:
            name: Model name to resolve into an existing model.
            unknown_name_mask: How to deal with unmatched names in name resolution, a mask of:

                - 2: Raise NameError if exact name not found (no generics)
                - 1: Only allow versioned names - raise NameError if generic non-versioned model name used
                - 0: Accept any name that can be resolved from OpenAIModel.known_models()

            genconf: Model generation configuration. Defaults to None.
            tokenizer: An external initialized tokenizer to use instead of the created from the GGUF file. Defaults to None.
            ctx_len: Maximum context length to be used (shared for input and output). Defaults to 0 which means model's maximum.
            api_key: OpenAI API key. Defaults to None, which will use env variable OPENAI_API_KEY.
            base_url: Base location for API access. Defaults to None, which will use env variable OPENAI_BASE_URL.
            openai_init_kwargs: Extra args for OpenAI.OpenAI() initialization. Defaults to {}.

        Raises:
            ImportError: If OpenAI API is not installed.
        """


        if not has_openai:
            raise ImportError("Please install openai by running: pip install openai")

        self._model_name, max_ctx_len, self._tokens_per_message, self._tokens_per_name = resolve_model(
            name,
            unknown_name_mask
        )

           
        super().__init__(False,
                         genconf,
                         schemaconf,
                         tokenizer
                         )

        # only check for "json" text presence as json schema is requested with the tools facility.
        self.json_format_instructors["json_schema"] = self.json_format_instructors["json"]
        
        logger.debug(f"Creating inner OpenAI with base_url={base_url}, openai_init_kwargs={openai_init_kwargs}")

        self._client = openai.OpenAI(api_key=api_key,
                                     base_url=base_url,
                        
                                     **openai_init_kwargs
                                     )

        
        # correct super __init__ values
        if self.tokenizer is None:
            self.tokenizer = OpenAITokenizer(self._model_name)

        if ctx_len == 0:
            self._ctx_len = max_ctx_len
        else:
            self._ctx_len = ctx_len
        

        self.TOOLS_TOKEN_LEN_FACTOR = self.DEFAULT_TOOLS_TOKEN_LEN_FACTOR




    
    def gen(self, 
            thread: Thread,
            genconf: Optional[GenConf] = None,
            ) -> GenOut:
        """Text generation from a Thread, used by the other model generation methods.
        Doesn't raise an exception if an error occurs, always returns GenOut.

        Args:
            thread: The Thread to use as model input.
            genconf: Model generation configuration. Defaults to None.

        Raises:
            NotImplementedError: If method was not defined by a derived class.

        Returns:
            A GenOut object with result, generated text, etc.
            The output text is in GenOut.text.
        """

        if genconf is None:
            genconf = self.genconf

        thread = self._prepare_gen_in(thread, genconf)

        token_len = self.token_len(thread, genconf)
        if genconf.max_tokens == 0:
            max_tokens = self.ctx_len - token_len            
            genconf = genconf(max_tokens=max_tokens)
            
        elif token_len + genconf.max_tokens > self.ctx_len:
            # this is not true for all models: 1106 models have 128k max input and 4k max output (in and out ctx are not shared)
            # so we assume the smaller max ctx length for the model
            logger.warn(f"Token length + genconf.max_tokens ({token_len + genconf.max_tokens}) is greater than model's context window length ({self.ctx_len})")

        fn_name = "json_out"

        json_kwargs: dict = {}
        format = genconf.format
        if format == "json":
            
            if genconf.json_schema is None:
                json_kwargs["response_format"] = {"type": "json_object"}

            else:
                # use json_schema in OpenAi's tool API
                json_kwargs["tool_choice"] = {
                    "type": "function",
                    "function": {"name": fn_name},
                }

                if isinstance(genconf.json_schema, str):
                    params = json.loads(genconf.json_schema)
                else:
                    params = genconf.json_schema
                
                json_kwargs["tools"] = [
                    {
                        "type": "function",
                        "function": {
                            "name": fn_name,
                            "parameters": params
                        }
                    }
                ]

        logger.debug(f"OpenAI json args: {json_kwargs}")
        
        msgs = thread.as_chatml()

        # https://platform.openai.com/docs/api-reference/chat/create
        response = self._client.chat.completions.create(model=self._model_name,
                                                        messages=msgs, # type: ignore[arg-type]
                                                        
                                                        max_tokens=genconf.max_tokens,
                                                        stop=genconf.stop,
                                                        temperature=genconf.temperature,
                                                        top_p=genconf.top_p,
                                                        **json_kwargs,
                                            
                                                        n=1
                                                        )

        logger.debug(f"OpenAI response: {response}")

        choice = response.choices[0]
        finish = choice.finish_reason
        message = choice.message

        if "tool_choice" in json_kwargs:
            
            # json schema generation via the tools API:
            if message.tool_calls is not None:
                fn = message.tool_calls[0].function
                if fn.name != fn_name:
                    logger.debug(f"OpenAIModel: different returned JSON function name ({fn.name})")

                text = fn.arguments
            else: # use content instead
                text = message.content # type: ignore[assignment]
        
        else:
            # text or simple json format
            text = message.content # type: ignore[assignment]

        out = self._prepare_gen_out(text, finish, genconf)

        return out

        


    


    def token_len(self,
                  thread: Thread,
                  genconf: Optional[GenConf] = None) -> int:
        """Calculate the number of tokens used by a list of messages.
        If a json_schema is provided in genconf, we use its string's token_len as upper bound for the extra prompt tokens.
        
        From https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        
        More info on calculating function_call (and tools?) tokens:

        https://community.openai.com/t/how-to-calculate-the-tokens-when-using-function-call/266573/24
        
        https://gist.github.com/CGamesPlay/dd4f108f27e2eec145eedf5c717318f5

        Args:
            thread: For token length calculation.
            genconf: Model generation configuration. Defaults to None.

        Returns:
            Estimated number of tokens the thread will use.
        """

        # name = self._model_name

        num_tokens = 0
        for index in range(-1, len(thread)): # -1 for system message
            message = thread.msg_as_chatml(index)
            # print(message)
            num_tokens += self._tokens_per_message
            for key, value in message.items():
                num_tokens += len(self.tokenizer.encode(value))
                # if key == "name":
                #     num_tokens += self._tokens_per_name
        
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        num_tokens += 10 # match API return counts

        # print("text token_len", num_tokens)

        if genconf is not None and genconf.json_schema is not None:
            if isinstance(genconf.json_schema, str):
                js_str = genconf.json_schema
            else:
                js_str = json.dumps(genconf.json_schema)

            tools_num_tokens = self.tokenizer.token_len(js_str)

            # this is an upper bound, as empirically tested with the api.
            tools_num_tokens = int(tools_num_tokens * self.TOOLS_TOKEN_LEN_FACTOR)

            # print("tools token_len", tools_num_tokens)

            num_tokens += tools_num_tokens
        
        return num_tokens



    
    @property
    def desc(self) -> str:
        """Model description."""
        return f"OpenAIModel: {self._model_name}"


    @classmethod
    def provider_version(_) -> str:
        """Provider library version: provider x.y.z
        Ex. openai 1.3.6
        """
        try:        
            import openai
            ver = openai.__version__
        except Exception:
            raise ImportError("Please install openai by running: pip install openai")
            
        return f"openai {ver}"
    
    @classmethod
    def known_models(cls) -> Union[list[str], None]:
        """If the model can only use a fixed set of models, return their names. Otherwise, return None.

        Returns:
            Returns a list of known models or None if it can accept any model.
        """
        return list(KNOWN_MODELS.keys())






class OpenAITokenizer(Tokenizer):
    """Tokenizer for OpenAI models."""

    def __init__(self, 
                 model: str
                 ):

        if not has_tiktoken:
            raise Exception("Please install tiktoken by running: pip install tiktoken")
        
        self._tok = tiktoken.encoding_for_model(model)

        self.vocab_size = self._tok.n_vocab
        
        self.bos_token_id = None
        self.bos_token = None

        self.eos_token_id = None
        self.eos_token = None

        self.pad_token_id = None
        self.pad_token = None

        self.unk_token_id = None
        self.unk_token = None
  

    
    def encode(self, 
               text: str) -> list[int]:
        """Encode text into model tokens. Inverse of Decode().

        Args:
            text: Text to be encoded.

        Returns:
            A list of ints with the encoded tokens.
        """
        return self._tok.encode(text)

        
    def decode(self, 
               token_ids: list[int],
               skip_special: bool = True) -> str:
        """Decode model tokens to text. Inverse of Encode().

        Args:
            token_ids: List of model tokens.
            skip_special: Don't decode special tokens like bos and eos. Defaults to True.

        Returns:
            Decoded text.
        """
        assert skip_special, "OpenAITokenizer only supports skip_special=True"

        return self._tok.decode(token_ids)

   




